package nl.kallestruik.dlib.gui

import kotlinx.coroutines.runBlocking
import net.kyori.adventure.text.Component
import org.bukkit.Bukkit
import org.bukkit.Material
import org.bukkit.entity.Player
import org.bukkit.event.EventHandler
import org.bukkit.event.HandlerList
import org.bukkit.event.Listener
import org.bukkit.event.inventory.ClickType
import org.bukkit.event.inventory.InventoryAction
import org.bukkit.event.inventory.InventoryClickEvent
import org.bukkit.event.inventory.InventoryCloseEvent
import org.bukkit.inventory.InventoryView
import org.bukkit.inventory.ItemStack
import org.bukkit.plugin.Plugin

fun chestGUI(player: Player, plugin: Plugin, configure: suspend ChestGUI.() -> Unit) {
    val chestGUI = ChestGUI()
    runBlocking {
        chestGUI.configure()
    }

    chestGUI.display(player, plugin)
}

class ChestGUI: Listener {
    private data class Configuration(
        var border: (suspend ChestGUIEntry.() -> Unit)? = null,
        var content: MutableList<suspend ChestGUIEntry.() -> Unit> = mutableListOf(),
        var itemStacks: MutableMap<Int, ItemStack> = mutableMapOf(),
        var onClose: suspend (gui: ChestGUI, player: Player, event: InventoryCloseEvent) -> Unit = { _, _, _ ->}
    )

    private val configuration = Configuration()
    var title: Component = Component.empty()
    var rows: Int = 0
    var cancelInteraction: Boolean = true

    private lateinit var view: InventoryView
    private lateinit var entries: Array<ChestGUIEntry?>


    fun onClose(onClose: suspend (gui: ChestGUI, player: Player, event: InventoryCloseEvent) -> Unit) {
        configuration.onClose = onClose
    }

    fun entry(entry: suspend ChestGUIEntry.() -> Unit) {
        configuration.content.add(entry)
    }

    fun border(border: suspend ChestGUIEntry.() -> Unit) {
        configuration.border = border
    }

    fun itemStack(slot: Int, itemStack: ItemStack) {
        configuration.itemStacks[slot] = itemStack
    }

    fun display(player: Player, plugin: Plugin) {
        val inventory = Bukkit.createInventory(null, rows * 9, title)
        view = player.openInventory(inventory) ?: return

        entries = arrayOfNulls(rows * 9)

        if (configuration.border != null) {
            val border = ChestGUIEntry()
            runBlocking {
                configuration.border!!.invoke(border)
            }

            val borderStack = border.toItemStack()

            // Top/Bottom border
            for (x in 0..8) {
                view.setItem(x, borderStack)
                view.setItem((rows - 1) * 9 + x, borderStack)
            }

            // Middle side border
            for (y in 1..rows - 2) {
                view.setItem(9 * y, borderStack)
                view.setItem(9 * y + 8, borderStack)
            }
        }

        configuration.content.forEach {
            val entry = ChestGUIEntry()
            runBlocking {
                it.invoke(entry)
            }

            if (entry.getSlot() < 0)
                return@forEach

            entries[entry.getSlot()] = entry
        }

        entries.forEach { entry ->
            if (entry == null)
                return@forEach

            if (entry.x != -1 || entry.y != -1)
                return@forEach

            inventory.setItem(entry.getSlot(), entry.toItemStack())
        }

        configuration.itemStacks.forEach { (slot, itemStack) ->
            inventory.setItem(slot, itemStack)
        }

        plugin.server.pluginManager.registerEvents(this, plugin)
    }

    private fun cleanup() {
        HandlerList.unregisterAll(this)
    }

    /*
    Bukkit event handlers
     */
    @EventHandler
    fun onInventoryClose(event: InventoryCloseEvent) {
        if (event.view != this.view)
            return

        cleanup()

        if (event.reason == InventoryCloseEvent.Reason.PLUGIN) {
            return
        }

        if (event.player !is Player)
            return

        runBlocking {
            configuration.onClose(this@ChestGUI, event.player as Player, event)
        }
    }

    @EventHandler
    fun onClick(event: InventoryClickEvent) {
        if (event.view != view) {
            return
        }

        if (event.action != InventoryAction.PICKUP_ALL) {
            if (cancelInteraction)
                event.isCancelled = true
            return
        }

        val slot = view.convertSlot(event.rawSlot)

        if (cancelInteraction)
            event.isCancelled = true

        val entry = entries[slot] ?: return

        if (event.whoClicked !is Player)
            return

        runBlocking {
            when (event.click) {
                ClickType.LEFT -> entry.onLeftClick(event)
                ClickType.MIDDLE -> entry.onMiddleClick(event)
                ClickType.RIGHT -> entry.onRightClick(event)
                else -> entry.onOtherClick(event)
            }
        }
    }
}

data class ChestGUIEntry(
    var name: Component = Component.empty(),
    var material: Material = Material.AIR,
    var amount: Int = 1,
    var lore: MutableList<Component> = mutableListOf(),
    var x: Int = -1,
    var y: Int = -1,
    var onLeftClick: suspend (event: InventoryClickEvent) -> Unit = {},
    var onMiddleClick: suspend (event: InventoryClickEvent) -> Unit = {},
    var onRightClick: suspend (event: InventoryClickEvent) -> Unit = {},
    var onOtherClick: suspend (event: InventoryClickEvent) -> Unit = {},
) {

    fun onLeftClick(onLeftClick: suspend (event: InventoryClickEvent) -> Unit) {
        this.onLeftClick = onLeftClick
    }

    fun onMiddleClick(onMiddleClick: suspend (event: InventoryClickEvent) -> Unit) {
        this.onMiddleClick = onMiddleClick
    }

    fun onRightClick(onRightClick: suspend (event: InventoryClickEvent) -> Unit) {
        this.onRightClick = onRightClick
    }

    fun onOtherClick(onOtherClick: suspend (event: InventoryClickEvent) -> Unit) {
        this.onOtherClick = onOtherClick
    }

    fun lore(line: Component) {
        lore.add(line)
    }

    fun getSlot(): Int {
        return x + y * 9
    }

    fun toItemStack(): ItemStack {
        val itemStack = ItemStack(material, amount)
        val itemMeta = itemStack.itemMeta
        itemMeta.displayName(name)
        itemMeta.lore(lore)

        itemStack.itemMeta = itemMeta
        return itemStack
    }
}